import requests
try:
    from bs4 import BeautifulSoup
except:
    print('you need to install BeautifulSoup4\n$ : python -m pip install BeautifulSoup4')

country_list = [
    '미국', '유럽연합', '일본', '중국', '홍콩', '대만', '영국', '오만', '캐나다', 
    '스위스', '스웨덴', '호주', '뉴질랜드', '체코', '칠레', '터키', '몽골', '이스라엘', 
    '덴마크', '노르웨이', '사우디아라비아', '쿠웨이트', '바레인', '아랍에미리트', '요르단', 
    '이집트', '태국', '싱가포르', '말레이시아', '인도네시아', '카타르', '카자흐스탄', '브루나이', 
    '인도', '파키스탄', '방글라데시', '필리핀', '멕시코', '브라질', '베트남', '남아프리카', 
    '러시아', '헝가리', '폴란드', '스리랑카', '알제리', '케냐', '콜롬비아', '탄자니아', 
    '네팔', '루마니아', '리비아', '마카오', '미얀마', '에티오피아', '우즈베키스탄', '캄보디아', '피지'
]

symbols = [
'AUD', 'USD', 'JPY', 'CNY', 'HKD', 'TWD', 'NTD', 'GBP', 'OMR', 'CAD', 'CHF', 'SEK', 'AUD', 'NZD',
'CZK', 'CLP', 'TRY', 'MNT', 'ILS', 'DKK', 'NOK', 'SAR', 'KWD', 'BHD', 'AED', 'JOD', 'EGP', 'THB',
'SGD', 'MYR', 'IDR', 'QAR', 'KZT', 'BND', 'INR', 'PKR', 'BDT', 'PHP', 'MXN', 'BRL', 'VND', 'ZAR',
'RUB', 'HUF', 'PLN', 'LKR', 'DZD', 'KES', 'COP', 'TZS', 'NPR', 'RON', 'LYD', 'MOP', 'MMK', 'ETB',
'UZS', 'KHR', 'FJD',
]

name_to_symbol = {'미국': 'USD', '유럽연합': 'EUR', '일본': 'JPY', '중국': 'CNY', '홍콩': 'HKD',
                  '대만': 'NTD', '영국': 'GBP', '오만': 'OMR', '캐나다': 'CAD', '스위스': 'CHF',
                  '스웨덴': 'SEK', '호주': 'AUD', '뉴질랜드': 'NZD', '체코': 'CZK', '칠레': 'CLP',
                  '튀르키예': 'TRY', '몽골': 'MNT', '이스라엘': 'ILS', '덴마크': 'DKK', '노르웨이': 'NOK',
                  '사우디아라비아': 'SAR', '쿠웨이트': 'KWD', '바레인': 'BHD', '아랍에미리트': 'AED',
                  '요르단': 'JOD', '이집트': 'EGP', '태국': 'THB', '싱가포르': 'SGD', '말레이시아': 'MYR', 
                  '인도네시아': 'IDR', '카타르': 'QAR', '카자흐스탄': 'KZT', '브루나이': 'BND', '인도': 'INR',
                  '파키스탄': 'PKR', '방글라데시': 'BDT', '필리핀': 'PHP', '멕시코': 'MXN', '브라질': 'BRL',
                  '베트남': 'VND', '남아프리카 공화국': 'ZAR', '러시아': 'RUB', '헝가리': 'HUF', '폴란드': 'PLN',
                  '스리랑카': 'LKR', '알제리': 'DZD', '케냐': 'KES', '콜롬비아': 'COP', '탄자니아': 'TZS',
                  '네팔': 'NPR', '루마니아': 'RON', '리비아': 'LYD', '마카오': 'MOP', '미얀마': 'MMK', 
                  '에티오피아': 'ETB', '우즈베키스탄': 'UZS', '캄보디아': 'KHR', '피지': 'FJD'}

symbol_to_name = {
    v: k
    for k, v in name_to_symbol.items()
}

def get_currency(symbol):
    symbol = symbol.upper()
    if symbol == 'KRW'or symbol =='한국':
        return 1.0
    
    assert symbol in symbol_to_name or symbol in name_to_symbol, '유효하지 않은 symbol 입니다.'

    if symbol in symbol_to_name:
        name = symbol_to_name[symbol]
    else:
        name = name_to_symbol[symbol]
    r = requests.get(f'https://www.google.com/search?q={name}환율')
    sp = BeautifulSoup(r.text, 'html.parser')
    return float(sp.select_one('.BNeawe.iBp4i.AP7Wnd').text.split()[0].replace(',', ''))



from typing import Union
import requests
import httpx

class CookieManager:

    _cookies:dict
    
    def __init__(self, cookie_path:str='./cookie') -> None:
        self._cookie_path = cookie_path
        self._cookies = {}

        try:
            open(cookie_path, 'r')
            self._load_cookies()
        except: pass

    def __del__(self):
        self._save_cookies()
        
    def add_cookies(self, cookies:dict) -> None:
        self._cookies.update(cookies)

    def get_cookie(self, k:str) -> str:
        return self._cookies.get(k, '')

    def _set_cookies(self, r:Union[requests.Response, httpx.Response]) -> None:
        if isinstance(r, requests.Response):
            self._cookies.update(r.cookies.get_dict())
        elif isinstance(r, httpx.Response):
            self._cookies.update(dict(r.cookies))
        ...

    def _reset_cookies(self, keys:list=[]) -> None:
        self._cookies = {k: self._cookies[k] for k in keys}
    
    def _save_cookies(self) -> None:
        try:
            open(self._cookie_path, 'w').write('\n'.join(['{}={}'.format(k, v) for k, v in self._cookies.items()]))
        except:
            ...
        
    def _load_cookies(self) -> dict:
        for l in open(self._cookie_path, 'r').readlines():
            _idx = l.index('=')
            self._cookies.update({l[:_idx]: l[_idx+1:].rstrip()})


import os
import email
import imaplib
import smtplib

from email import encoders
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header import decode_header

from os.path import basename

import time
from datetime import datetime

import traceback
from typing import Union

from dataclasses import dataclass

def _is_ascii(s):
    '''
    ASCII 체크
    '''
    return all(ord(c) < 128 for c in s)

def _get_subject(subject):
    '''
    메일 제목 반환
    '''
    s, encoding = decode_header(subject)[0]
    if encoding:
        return str(s, encoding, 'replace')
    else:
        return s

def _get_body(msg):
    '''
    메일 내용 반환
    '''
    if msg.is_multipart():
        return _get_body(msg.get_payload(0))
    else:
        return msg.get_payload(None, True)

@dataclass
class EmailAccount:
    email_id: str
    email_pw: str

class EmailManager:
    """
    Email Manager
    """

    def __init__(self, email_id, email_pw, mail_server:str='naver', mode='send') -> None:
        """
        mail_server: gmail, naver
        mode: send, get
        """
        self.email_id = email_id
        self.email_pw = email_pw
        
        if mode == 'send':
            SMTP_SERVER = f'smtp.{mail_server.lower()}.com'
            SMTP_PORT = 465
            smtp = smtplib.SMTP_SSL(SMTP_SERVER, SMTP_PORT)
            smtp.login(email_id, email_pw)
            self.protocol = smtp
        elif mode == 'get':
            IMAP_SERVER = f"imap.{mail_server.lower().strip()}.com" 
            IMAP_PORT = 993
            imap = imaplib.IMAP4_SSL(IMAP_SERVER, IMAP_PORT)
            imap.login(self.email_id, self.email_pw)
            self.protocol = imap

    def __del__(self) -> None:
        self.protocol.close()
        # SMTP_SSL has no method logout
        if type(self.protocol) != smtplib.SMTP_SSL:
            self.protocol.logout()

    def get_email(self, header_subjects:list=[], header_since:str=None, header_from:str=None, 
                  label:str=None, encode_type='utf-8', save_path:str=None, remove_email=True
                    ) -> list[dict]:
        """
        메일 제목에 특정 문자열이 있거나 특정 발송인의 이메일 반환\n
        **필수값**\n
        header_str : 검색할 제목 (non-ascii) | header_from : 발송인\n
        **선택값**\n
        save_path : 첨부 파일 저장 경로 (지정하지 않을 시 첨부파일 저장되지 않음)\n
        return list[dict]
        [
            {
                'subject': subject,
                'date': date
                'content': mail_content,
                'attaches': [path:str]
            }
        ]
        """

        # 메일 검색어 ASCII 체크
        if header_subjects:
            assert all([_is_ascii(header_str) for header_str in header_subjects]), 'In ASCII 문자열만 입력해주세요 (영문, 기호)'
        assert header_subjects or header_from or header_since or label, '메일 제목이나 발신인, 날짜, 라벨을 추가해주세요.'

        email_result = []

        try:
            self.protocol.select('inbox') if not label else self.protocol.select(label)
            search_list = []
            if header_subjects:
                if len(header_subjects) == 1:
                    search_list.append(f'(Header Subject "{header_subjects[0]}")')
                else:
                    # 230315 업데이트
                    search_list.append('OR ' * (len(header_subjects)-1) + ' '.join([f'Subject "{hs}"' for hs in header_subjects]) if header_subjects else '')
            
            if header_from:
                search_list.append(f'(from "{header_from}")' if header_from else '')
                
            if header_since:
                search_list.append(f'(SINCE "{header_since}")')

            if search_list:
                _, data = self.protocol.uid('search', None, *search_list)
            else:
                _, data = self.protocol.uid('search', None, 'All')

            for target_uid in data[0].split():
                try:
                    _, data = self.protocol.uid('fetch', target_uid, '(RFC822)')
                    try:
                        email_message = email.message_from_string(data[0][1].decode(encode_type))
                    except:
                        email_message = email.message_from_string(data[0][1].decode('cp949')) #euc-kr

                    # 메일 제목
                    subject = _get_subject(email_message['subject'])

                    # 메일 내용
                    try:
                        mail_content = _get_body(email_message).decode(encode_type).strip()
                    except:
                        mail_content = _get_body(email_message).decode('cp949').strip()

                    # 수신 일시
                    timestamp = (time.mktime(email.utils.parsedate(email_message['Date'])))
                    mail_time = datetime.fromtimestamp(timestamp)

                    # 첨부 파일
                    file_list = []
                    for part in email_message.walk():
                        if part.get_content_maintype() == 'multipart':
                            continue
                        if part.get('Content-Disposition') is None:
                            continue

                        try:
                            filename = part.get_filename()
                            s, encoding = decode_header(filename)[0]
                            if encoding:
                                filename = str(s, encoding, 'replace')

                            file_list.append({
                                'name': filename,
                                'file': part.get_payload(decode=True)
                            })
                        except:
                            pass

                    saved_files = []

                    # 첨부파일 저장여부
                    if save_path:
                        os.makedirs(save_path, exist_ok=True)
                        for eval_file in file_list:
                            att_path = os.path.join(save_path, eval_file['name'])
                            saved_files.append(str(att_path))
                            if not os.path.isfile(att_path):
                                with open(att_path, 'wb') as fp:
                                    fp.write(eval_file['file'])

                    # 이메일 지우기(지울 메세지의 uid)
                    if remove_email:
                        self.protocol.uid('store', target_uid, '+FLAGS', '(\\Deleted)')
                        # 메일함 정리하기
                        self.protocol.expunge()

                    email_result.append({
                        'subject':subject,
                        'date': mail_time,
                        'content': mail_content,
                        'attaches': saved_files,
                    })
                except:
                    with open('./email_error', 'a') as f:
                        f.write('에러 발생, ' + str(target_uid) + '\n')

        except:
            traceback.print_exc()
            
        return email_result
    
    
    # 이메일 발송 함수
    def send_email(self, from_user:str, to_users:list, 
                   subject:str, contents:list[Union[tuple, str]],
                   attachments:list=[], cc_targets=[], is_html=False,
                   ) -> bool:
        '''
        메일을 발송하는 함수입니다.

        **필수값**
        from_user: 보내는 사람의 이메일 주소
        to_users: 받는 사람의 이메일 주소들 (list)
        subject: 메일 제목
        content: 메일 내용 (list)

        **선택**
        attachments: 첨부 파일들 (파일경로 리스트)
        cc_targets: 참조 이메일 주소들 (list)
        '''

        try:
            msg = MIMEMultipart('alternative')

            if attachments:
                msg = MIMEMultipart('mixed')

                for attachment in attachments:
                    email_file = MIMEBase('application', 'octet-stream')

                    with open(attachment, 'rb') as f:
                        file_data = f.read()

                    email_file.set_payload(file_data)
                    encoders.encode_base64(email_file)

                    file_name = basename(attachment)
                    email_file.add_header('Content-Disposition', 'attachment',
                        filename=file_name)

                    msg.attach(email_file)

            msg['From'] = from_user
            msg['To'] = ','.join(set(to_users))
            if cc_targets:
                msg['CC'] = ','.join(cc_targets)
            msg['Subject'] = subject

            # if type contents is string -> 552 5.2.3 error
            if type(contents) == str: contents = [contents]
            for content in contents:
                # don't care content is str or tuple
                msg.attach(MIMEText(content, 'html' if is_html else 'plain'))

            self.protocol.sendmail(from_user, set(to_users+cc_targets), msg.as_string())
            return True

        except Exception:
            traceback.print_exc()
        
        return False

# deprecated
def _get_emails(email_id, email_pw, **kwargs):
    """
    kwargs
    mail_server = ['gmail', 'naver']
    header_subjects:list=[], header_since:str=None, header_from:str=None, 
    label:str=None, save_path:str=None, 

    remove_email=True,
    encode_type='utf-8',
    """
    emgr = EmailManager(email_id, email_pw, mail_server=kwargs.pop('mail_server', 'naver'), mode='get')
    emails = emgr.get_email(**kwargs)
    del emgr
    return emails

# deprecated
def _send_email(email_id, email_pw, **kwargs):
    """
    kwargs
    mail_server = ['gmail', 'naver']
    from_user:str, to_users:list, 
    subject:str, contents:list[Union[tuple, str]],
    attachments:list=[], cc_targets=[]
    """
    emgr = EmailManager(email_id, email_pw, mail_server=kwargs.pop('mail_server', 'naver'))
    result = emgr.send_email(**kwargs)
    del emgr
    return result

class GmailManager(EmailManager):
    def __init__(self, email_id, email_pw, mode='send') -> None:
        super().__init__(email_id, email_pw, 'gmail', mode)

class NaverManager(EmailManager):
    def __init__(self, email_id, email_pw, mode='send') -> None:
        super().__init__(email_id, email_pw, 'naver', mode)

def send_with_gmail(email_id, email_pw, **kwargs):
    return GmailManager(email_id, email_pw).send_email(**kwargs)

def send_with_naver(email_id, email_pw, **kwargs):
    return NaverManager(email_id, email_pw).send_email(**kwargs)

def get_with_gmail(email_id, email_pw, **kwargs):
    return GmailManager(email_id, email_pw).get_email(**kwargs)

def get_with_naver(email_id, email_pw, **kwargs):
    return NaverManager(email_id, email_pw).get_email(**kwargs)

if __name__ == '__main__':
    config = {
        'EMAIL_ID': 'your-email-id',
        'EMAIL_PW': 'your-email-pw',
    }
    emgr = EmailManager(config.get('EMAIL_ID'), config.get('EMAIL_PW'), mail_server='gmail')
    emgr.send_email(config['EMAIL_ID'], ['target-user-email'], 'this is test email !', 
                    [('this is test content', 'plain'), 'this is just text', ('<div style="color:red;">this is html</div>', 'html')])
    
    # date_now = datetime(2022, 3, 7).strftime("%d-%b-%Y")
    result = emgr.get_email(
        # header_subjects=['test', 'store'],
        # header_from='svstar94@gmail.com',
        # header_since=date_now,
        label='report',
        remove_email=False, 
    )


try:
    import psycopg2
    from psycopg2.extras import RealDictCursor
except:
    print('you need to install psycopg2\n$ : python -m pip install psycopg2')

from typing import Iterable

from sosin.databases.db import Database

class PostgreSQL(Database):

    def __init__(self, db_config, cursor_type:str = None):
        """
        cursor_type = 'dict' or None
        """
        db_config['port'] = int(db_config.get('port', '5432'))
        db_config['dbname'] = db_config.pop('database')
        db_config.pop('charset', None)

        if cursor_type == 'dict':
            cursor_type = RealDictCursor
        else:
            cursor_type = None

        super().__init__(db_config, cursor_type)

    def connect(self):
        self.DB = psycopg2.connect(**self.db_config)
        self.cursor = self.DB.cursor(self.cursor_type)
        return self
    
    def custom_select(self, qry):
        return super().select(qry)
    
    def select(self, table, columns, where_info=None, order_by=None, page_info=None):
        """
        table -> should be schema_name.table_name
        """
        query = ''

        if not isinstance(columns, str) and isinstance(columns, Iterable):
            columns = ', '.join(map(str, columns))
        
        query += f'SELECT {columns} FROM {table}'

        if where_info:
            query += self.where(*where_info)

        if order_by:
            query += self.sort(order_by)

        if page_info:
            query += self.pagination(*page_info)

        return super().select(query)
    
    # where equal
    # where between
    # where in
    # where null / not null
    def where(self, column, value, option=None): 
        """
        option IN, IS, ...
        BETWEEN A AND B (VALUES)

        """
        if not option:
            option = '='

        return f'\nWHERE {column} {option} {value}'

    def sort(self, order_by):
        return f"\nORDER BY {order_by}"
    
    def pagination(self, page, size):
        return f'\nLIMIT {size} OFFSET {(page-1)*size}'

    def insert(self, table, columns, value) -> bool:
        if not isinstance(columns, str) and isinstance(columns, Iterable):
            columns = ', '.join(map(str, columns))
        
        assert len(value) == columns.count(',')+1, "칼럼 길이와 값의 길이가 다릅니다. ,를 확인해주세요"

        value_qry = ','.join(['%s'] * len(value))
        query = f'INSERT INTO {table}({columns}) VALUES ({value_qry});'

        return super().insert(query, value)
    
    def insert_many(self, table, columns, values):
        if not isinstance(columns, str) and isinstance(columns, Iterable):
            columns = ', '.join(map(str, columns))
        
        assert len(values[0]) == columns.count(',')+1, "칼럼 길이와 값의 길이가 다릅니다. ,를 확인해주세요"

        value_qry = ','.join(['%s'] * len(values[0]))
        query = f'INSERT INTO {table}({columns}) VALUES ({value_qry});'

        return super().insert_many(query, values)
    
    def update(self, table, columns, values, where_info) -> bool:
        assert len(columns) == len(values), "칼럼 길이와 값의 길이가 다릅니다."

        query = f'UPDATE {table}\nSET '
        set_query = ',\n'.join(f'{column}=%s' for column in columns)
        query += set_query

        # where 도 %s로 안전하게?
        query += self.where(*where_info)
        query += ';'
        return super().update(query, values)

    def delete(self, table, where_info):
        query = f'DELETE FROM {table}'
        query += self.where(*where_info)
        query += ';'
        return super().delete(query)

if __name__ == '__main__':
    db = PostgreSQL({'host': '3.39.222.13', 'port': '5432', 'user':'tready', 'password': 'test', 'database': 'tready_db'})

    with db:
        result = db.select('tb_store', '*')

    print(result)



class HistoryManager:

    history: list
    
    def __init__(self) -> None:
        self.history = []
    
    def add_history(self, r):
        self.history.append(r)

    def get_histories(self):
        return self.history